﻿// Copyright (C) 2011 Oliver Sturm <oliver@oliversturm.com>. All rights reserved.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.CompilerServices;
using System.Diagnostics;

namespace AsyncAwaitMyself {
  class Program {
    static void Main(string[] args) {
      MyClassTest( );
      TimeSpanTest( );
      NotepadTest( );

      Console.ReadLine( );
    }

    async static void MyClassTest( ) {
      Console.WriteLine("Starting MyClassTest");

      // Never forget how the whole thing works!!! The following line does 
      // NOT mean to wait for completion of the class, or anything like that. 
      // Instead, it only means that the code that comes after this line
      // will only be executed once the work represented by MyClass has
      // been completed.
      var result = await new MyClass( );

      Console.WriteLine("Done with MyClassTest - result is " + result);
    }

    async static void TimeSpanTest( ) {
      Console.WriteLine("Starting TimeSpanTest");

      await TimeSpan.FromSeconds(3);

      Console.WriteLine("End of TimeSpanTest");
    }

    async static void NotepadTest( ) {
      Console.WriteLine("Starting NotepadTest");

      await Process.Start("notepad.exe");

      Console.WriteLine("End of NotepadTest");
    }


  }

  public static class Extensions {
    // Thanks to Stephen Toub for these samples!

    public static TaskAwaiter GetAwaiter(this TimeSpan ts) {
      return Task.Delay(ts).GetAwaiter();
    }

    public static TaskAwaiter<int> GetAwaiter(this Process process) {
      var tcs = new TaskCompletionSource<int>( );
      process.EnableRaisingEvents = true;
      process.Exited += (s, e) => tcs.TrySetResult(process.ExitCode);
      if (process.HasExited)
        tcs.TrySetResult(process.ExitCode);
      return tcs.Task.GetAwaiter( );
    }
  }

  class MyClass {
    public MyClassAwaiter GetAwaiter( ) {
      Console.WriteLine("GetAwaiter");
      return new MyClassAwaiter();
    }

    public class MyClassAwaiter {
      public bool IsCompleted {
        get {
          // Important: saying "no" here doesn't mean you're asked a second time!
          // It is just used to detect whether the task is perhaps done already,
          // so the runtime doesn't have to do the whole thing of setting up the 
          // continuation etc etc.
          Console.WriteLine("Am I done?? No, not yet.");
          return false;
        }
      }

      public void OnCompleted(Action continuation) {
        Console.WriteLine("In OnCompleted, now executing action...");

        continuation();

        Console.WriteLine("action done.");
      }

      public int GetResult() {
        return 42;
      }
    }
  }
}
